#!perl

#
# Módulo de prueba contra regresiones en salida
# Antonio García Domínguez
# $Id$
#
use strict;
use warnings;
use English;

use Test::More;
use XML::Validate;
use XML::SemanticDiff;
use XML::Parser;

use Cwd;
use File::Copy;
use File::Spec;

use ACL2::Proyecto;
use ACL2::Test::ManejDetalleDiff;

# Nombres de los documentos a usar para pruebas
my @docsPrueba = map { m!.*/(.*?)\.lisp!i } <t/testInputs/*.lisp>;

# Avisamos a Test::More del número de pruebas a realizar
plan tests => 2 * scalar @docsPrueba;

# Generamos el validador de XML
my $validador = new XML::Validate(Type => 'BestAvailable');

foreach my $i (@docsPrueba) {
    my $fuenteLisp  = "t/testInputs/${i}.lisp";
    my $fuenteDem   = "t/testInputs/${i}.lisp.out";
    my $salidaXML   = "t/testInputs/${i}.lisp.xml";
    my $esperadoXML = "t/expectedOutputs/${i}.xml.test";

    my $demo = new ACL2::Proyecto("ruta_lisp" => $fuenteLisp);
    $demo->actualizar('force' => 1);
    my $cadenaXML = volcarFichero($salidaXML);

    ##################################################
    # Validamos la salida XML contra la DTD incrustada
    ##################################################

    my $valido = $validador->validate($cadenaXML);
    ok($valido, "${salidaXML} must be well-formed and valid according to its DTD");
    if (!$valido) {
	print "VALIDATION ERRORS:\n" .
	    "------------------------\n";

	while (my ($clave, $contenido) = each %{$validador->last_error()}) 
	{
	    print "${clave}: $contenido\n";
	}
	
	print "------------------------\n";
    }

    ###################################################
    # Comprobamos que la salida es la esperada
    ###################################################

    # Si no existe la salida esperada, asumimos que es la primera vez
    # que se ejecuta el test y copiamos la obtenida
    if (! -e $esperadoXML) {
	copy($salidaXML, $esperadoXML) 
	    or die "Couldn't copy $salidaXML to $esperadoXML: $!";
    }

    my $cadEsperadoXML  = volcarFichero($esperadoXML);
    my $hay_cambios_imp = compararSalidas($cadenaXML, $cadEsperadoXML);
    ok($hay_cambios_imp == 0,  
       "${salidaXML} should have the expected contents");

}

=head2 volcarFichero($ruta)

Uso: Devuelve una cadena con el contenido del fichero con la ruta relativa $ruta.

=cut

sub volcarFichero {
    my ($ruta) = @_;

    open(F_VOLCARFICHERO, $ruta) or die "Couldn't open $ruta: $!";
    my $res = join '', <F_VOLCARFICHERO>;
    close(F_VOLCARFICHERO);

    return $res;
}

=head2 compararSalidas($doc1, $doc2)

Uso: Devuelve 1 si existen diferencias importantes entre las cadenas
con las salidas XML $doc1 y $doc2, o 0 de lo contrario.

=cut

sub compararSalidas {
    my ($cadenaXML, $cadEsperadoXML) = @_;
    my $diff = new XML::SemanticDiff(
	keepdata => 1,
	diffhandler  => new ACL2::Test::ManejDetalleDiff( keepdata => 1 )
    );
    my $hay_cambios_importantes = 0;
    
    foreach my $change ($diff->compare($cadEsperadoXML, $cadenaXML) )
    {
	my ($event, $elem, $attr, $context) = ($change->{event}, 
					       $change->{element_name}, 
					       $change->{attribute_name},
					       $change->{context});
	
	# Filter unimportant events
	unless ($event eq 'attribute_value' 
		&& ($elem eq 'time' 
                    || $elem eq 'script' 
                    || $elem eq 'acl2_version' 
                    || $attr eq 'name' && $context =~ m!certifybook!i
		    || ($attr eq 'book_dir'  && same_path($change->{old_value}, $change->{new_value}))
		    || ($attr eq 'book_file' && same_path($change->{old_value}, $change->{new_value}))
		    || ($attr eq 'path_book' && same_path($change->{old_value}, $change->{new_value}))
		    || ($attr eq 'path_cert' && same_path($change->{old_value}, $change->{new_value}))
                    || normalize_whitespace($change->{old_value}) eq normalize_whitespace($change->{new_value}))
		|| $event eq 'element_value'
		&& $context =~ /paragraph\[[0-9]+\]$/i
		|| $context =~ m!certifybook.*/paragraph\[1\]!i
		|| $context =~ m!certifybook.*/compilation.*/paragraph\[2\]!i
		|| $context =~ m!certifybook.*/write_cert.*/sexpr\[1\]!i
		|| $context =~ m!includebook.*/output!i)
	{
	    print "$change->{message}: context $change->{context}\n";

	    if ($event eq 'attribute_value' || $event eq 'element_value') {
		print "Expected:\n" 
		    . $change->{old_value}
		    . "\nGot:\n" . $change->{new_value} . "\n";
	    }
	    
	    $hay_cambios_importantes = 1;
	}
    }

    return $hay_cambios_importantes;
}

sub normalize_whitespace {
    my ($a) = @_;

    $a =~ s/^\s+//g;
    $a =~ s/\s+$//g;
    $a =~ s/\s{2,}/ /g;

    return $a;
}

# Compares two paths and checks whether they are really the same. In
# Windows, it normalizes both paths to use forward slashes and lower
# case.
sub same_path {
    my ($a, $b) = @_;

    if ($OSNAME =~ /^(?:MSWin|cygwin)/) {
	$a =~ s#\\#/#g;
	$a = lc $a;
	$b =~ s#\\#/#g;
	$b = lc $b;
    }

    return $a eq $b;
}
